# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import tempfile
from typing import List
from unittest.mock import AsyncMock, patch
from pathlib import PurePath
import pytest
from itertools import cycle

from servicex.configuration import Configuration
from servicex.dataset_identifier import FileListDataset
from servicex.expandable_progress import ExpandableProgress
from servicex.func_adl.func_adl_dataset import FuncADLQuery_Uproot
from servicex.models import (
    TransformStatus,
    Status,
    ResultFile,
    ResultFormat,
    TransformRequest,
    TransformedResults,
)
from servicex.query_cache import QueryCache
from servicex.query_core import ServiceXException, Query
from servicex.servicex_client import ServiceXClient
from servicex.uproot_raw.uproot_raw import UprootRawQuery

transform_status = TransformStatus(
    **{
        "request_id": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf",
        "did": "File List Provided in Request",
        "columns": None,
        "selection": "(Where (SelectMany (call EventDataset) (lambda (list e) (call (attr e 'Jets') 'AntiKt4EMTopoJets'))) (lambda (list j) (and (> (/ (call (attr j 'pt')) 1000) 20) (< (call abs (/ (call (attr j 'eta')) 1000)) 4.5))))",  # NOQA 501
        "tree-name": None,
        "image": "sslhep/servicex_func_adl_uproot_transformer:uproot4",
        "workers": None,
        "result-destination": "object-store",
        "result-format": "parquet",
        "workflow-name": "selection_codegen",
        "generated-code-cm": "b8c508d0-ccf2-4deb-a1f7-65c839eebabf-generated-source",
        "status": "Complete",
        "failure-info": None,
        "app-version": "develop",
        "code-gen-image": "sslhep/servicex_code_gen_func_adl_uproot:v1.2.0",
        "files": 0,
        "files-completed": 0,
        "files-failed": 0,
        "files-remaining": None,
        "submit-time": "2023-05-25T20:05:05.564137Z",
        "finish-time": None,
        "minio-endpoint": "minio.org:9000",
        "minio-secured": False,
        "minio-access-key": "miniouser",
        "minio-secret-key": "letmein",
        "log-url": "https://dummy/()",
    }
)

transform_status1 = transform_status.model_copy(
    update={
        "status": Status.running,
        "files_remaining": None,
        "files_completed": 0,
        "files": 0,
    }
)
transform_status2 = transform_status.model_copy(
    update={
        "status": Status.running,
        "files_remaining": 1,
        "files_completed": 1,
        "files": 2,
    }
)
transform_status3 = transform_status.model_copy(
    update={
        "status": Status.complete,
        "files_remaining": 0,
        "files_completed": 2,
        "files": 2,
    }
)
transform_status4 = transform_status.model_copy(
    update={
        "status": Status.canceled,
        "files_remaining": 1,
        "files_completed": 1,
        "files": 2,
    }
)
transform_status5 = transform_status.model_copy(
    update={
        "status": Status.fatal,
        "files_remaining": 0,
        "files_completed": 1,
        "files_failed": 1,
        "files": 2,
    }
)
transform_status6 = transform_status.model_copy(
    update={
        "status": Status.complete,
        "files_remaining": 0,
        "files_completed": 1,
        "files_failed": 1,
        "files": 2,
    }
)
transform_status4 = transform_status.model_copy(
    update={
        "status": Status.canceled,
        "files_remaining": 1,
        "files_completed": 1,
        "files": 2,
    }
)
transform_status5 = transform_status.model_copy(
    update={
        "status": Status.fatal,
        "files_remaining": 0,
        "files_completed": 1,
        "files_failed": 1,
        "files": 2,
    }
)
transform_status6 = transform_status.model_copy(
    update={
        "status": Status.complete,
        "files_remaining": 0,
        "files_completed": 1,
        "files_failed": 1,
        "files": 2,
    }
)


@pytest.fixture
def servicex():
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    # servicex.get_transform_status.side_effect = [
    #     transform_status1,
    #     transform_status3,
    # ]
    yield servicex


file1 = ResultFile(filename="file1", size=100, extension="parquet")
file2 = ResultFile(filename="file2", size=100, extension="parquet")


def transformed_results(
    transform: TransformRequest,
    completed_status: TransformStatus,
    data_dir: str,
    file_list: List[str],
    signed_urls,
) -> TransformedResults:
    return TransformedResults(
        hash=transform.compute_hash(),
        title=transform.title,
        codegen=transform.codegen,
        request_id=completed_status.request_id,
        submit_time=completed_status.submit_time,
        data_dir=data_dir,
        file_list=file_list,
        signed_url_list=signed_urls,
        files=completed_status.files,
        result_format=transform.result_format,
        log_url=completed_status.log_url,
    )


def cache_transform(record: TransformedResults):
    return


@pytest.mark.asyncio
async def test_submit(mocker):
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status2,
        transform_status3,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1, file2]])
    mock_minio.download_file = AsyncMock(
        side_effect=lambda a, _, shorten_filename: PurePath(a)
    )

    mock_cache = mocker.MagicMock(QueryCache)
    mock_cache.get_transform_by_hash = mocker.MagicMock(return_value=None)
    mock_cache.transformed_results = mocker.MagicMock(side_effect=transformed_results)
    mock_cache.cache_transform = mocker.MagicMock(side_effect=cache_transform)
    mock_cache.cache_path_for_transform = mocker.MagicMock(return_value=PurePath("."))
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    datasource = Query(
        dataset_identifier=did,
        title="ServiceX Client",
        codegen="uproot",
        sx_adapter=servicex,
        query_cache=mock_cache,
        config=Configuration(api_endpoints=[]),
    )
    datasource.query_string_generator = FuncADLQuery_Uproot().FromTree("nominal")
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        result = await datasource.submit_and_download(
            signed_urls_only=False, expandable_progress=progress
        )
        print(mock_minio.download_file.call_args)
    assert result.file_list == ["file1", "file2"]
    mock_cache.cache_transform.assert_called_once()


@pytest.mark.asyncio
async def test_submit_partial_success(mocker):
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status2,
        transform_status6,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1]])
    mock_minio.download_file = AsyncMock(
        side_effect=lambda a, _, shorten_filename: PurePath(a)
    )

    mock_cache = mocker.MagicMock(QueryCache)
    mock_cache.get_transform_by_hash = mocker.MagicMock(return_value=None)
    mock_cache.transformed_results = mocker.MagicMock(side_effect=transformed_results)
    mock_cache.cache_transform = mocker.MagicMock(side_effect=cache_transform)
    mock_cache.cache_path_for_transform = mocker.MagicMock(return_value=PurePath("."))
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    datasource = Query(
        dataset_identifier=did,
        title="ServiceX Client",
        codegen="uproot",
        sx_adapter=servicex,
        query_cache=mock_cache,
        config=Configuration(api_endpoints=[]),
    )
    datasource.query_string_generator = FuncADLQuery_Uproot().FromTree("nominal")
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        result = await datasource.submit_and_download(
            signed_urls_only=False, expandable_progress=progress
        )
        print(mock_minio.download_file.call_args)
    assert result.file_list == ["file1"]
    mock_cache.cache_transform.assert_not_called()


@pytest.mark.asyncio
async def test_use_of_cache(mocker):
    """Do we pick up the cache on the second request for the same transform?"""
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status3,
    ]
    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(return_value=[file1, file2])
    mock_minio.download_file = AsyncMock(
        side_effect=lambda a, _, shorten_filename: PurePath(a)
    )
    mock_minio.get_signed_url = AsyncMock(side_effect=["http://file1", "http://file2"])

    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)

    did = FileListDataset("/foo/bar/baz.root")
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        datasource = Query(
            dataset_identifier=did,
            title="ServiceX Client",
            codegen="uproot",
            sx_adapter=servicex,
            query_cache=cache,
            config=config,
        )
        datasource.query_string_generator = FuncADLQuery_Uproot().FromTree("nominal")
        datasource.result_format = ResultFormat.parquet
        upd = mocker.patch.object(
            cache, "update_record", side_effect=cache.update_record
        )
        with ExpandableProgress(display_progress=False) as progress:
            result1 = await datasource.submit_and_download(
                signed_urls_only=True, expandable_progress=progress
            )
        upd.assert_not_called()
        upd.reset_mock()
        assert mock_minio.get_signed_url.await_count == 2
        # second round, should hit the cache (and not call the sx_adapter, minio, or update_record)
        with ExpandableProgress(display_progress=False) as progress:
            servicex2 = AsyncMock()
            mock_minio.list_bucket.reset_mock()
            mock_minio.get_signed_url.reset_mock()
            datasource2 = Query(
                dataset_identifier=did,
                title="ServiceX Client",
                codegen="uproot",
                sx_adapter=servicex2,
                query_cache=cache,
                config=config,
            )
            datasource2.query_string_generator = FuncADLQuery_Uproot().FromTree(
                "nominal"
            )
            datasource2.result_format = ResultFormat.parquet
            result2 = await datasource2.submit_and_download(
                signed_urls_only=True, expandable_progress=progress
            )
            servicex2.assert_not_awaited()
            mock_minio.list_bucket.assert_not_awaited()
            mock_minio.get_signed_url.assert_not_awaited()
        upd.assert_not_called()
        assert result1 == result2
        upd.reset_mock()
        servicex.get_transform_status.reset_mock(side_effect=True)
        servicex.get_transform_status.return_value = transform_status3
        mock_minio.list_bucket.reset_mock(side_effect=True)
        # third round, should hit the cache and download files (and call update_record)
        with ExpandableProgress(display_progress=False) as progress:
            await datasource.submit_and_download(
                signed_urls_only=False, expandable_progress=progress
            )
        servicex.assert_not_awaited()
        assert mock_minio.download_file.await_count == 2
        upd.assert_called_once()
        # fourth round, should hit the cache (and nothing else)
        mock_minio.list_bucket.reset_mock()
        mock_minio.download_file.reset_mock()
        with ExpandableProgress(display_progress=False) as progress:
            await datasource.submit_and_download(
                signed_urls_only=False, expandable_progress=progress
            )
        servicex.assert_not_awaited()
        mock_minio.list_bucket.assert_not_awaited()
        mock_minio.download_file.assert_not_awaited()
        upd.assert_called_once()
        cache.close()


@pytest.mark.asyncio
async def test_submit_cancel(mocker):
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status4,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1]])
    mock_minio.download_file = AsyncMock(
        side_effect=lambda a, _, shorten_filename: PurePath(a)
    )

    mock_cache = mocker.MagicMock(QueryCache)
    mock_cache.get_transform_by_hash = mocker.MagicMock(return_value=None)
    mock_cache.cache_transform = mocker.MagicMock(side_effect=cache_transform)
    mock_cache.cache_path_for_transform = mocker.MagicMock(return_value=PurePath("."))
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    datasource = Query(
        dataset_identifier=did,
        title="ServiceX Client",
        codegen="uproot",
        sx_adapter=servicex,
        query_cache=mock_cache,
        config=Configuration(api_endpoints=[]),
    )
    datasource.query_string_generator = FuncADLQuery_Uproot().FromTree("nominal")
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        with pytest.raises(ServiceXException):
            _ = await datasource.submit_and_download(
                signed_urls_only=False, expandable_progress=progress
            )
    mock_cache.cache_transform.assert_not_called()


@pytest.mark.asyncio
async def test_submit_fatal(mocker):
    servicex = AsyncMock()
    servicex.submit_transform = AsyncMock()
    servicex.submit_transform.return_value = {"request_id": '123-456-789"'}
    servicex.get_transform_status = AsyncMock()
    servicex.get_transform_status.side_effect = [
        transform_status1,
        transform_status5,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1]])
    mock_minio.download_file = AsyncMock(
        side_effect=lambda a, _, shorten_filename: PurePath(a)
    )

    mock_cache = mocker.MagicMock(QueryCache)
    mock_cache.get_transform_by_hash = mocker.MagicMock(return_value=None)
    mock_cache.cache_transform = mocker.MagicMock(side_effect=cache_transform)
    mock_cache.cache_path_for_transform = mocker.MagicMock(return_value=PurePath("."))
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    datasource = Query(
        dataset_identifier=did,
        title="ServiceX Client",
        codegen="uproot",
        sx_adapter=servicex,
        query_cache=mock_cache,
        config=Configuration(api_endpoints=[]),
    )
    datasource.query_string_generator = FuncADLQuery_Uproot().FromTree("nominal")
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        with pytest.raises(ServiceXException):
            _ = await datasource.submit_and_download(
                signed_urls_only=False, expandable_progress=progress
            )
    mock_cache.cache_transform.assert_not_called()


@pytest.mark.asyncio
async def test_submit_generic(mocker, codegen_list):
    """Uses Uproot-Raw classes which go through the generic query mechanism"""
    import json

    sx = AsyncMock()
    sx.submit_transform = AsyncMock()
    sx.submit_transform.return_value = {"request_id": '123-456-789"'}
    sx.get_transform_status = AsyncMock()
    sx.get_transform_status.side_effect = [
        transform_status1,
        transform_status2,
        transform_status3,
    ]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1, file2]])
    mock_minio.download_file = AsyncMock()

    mock_cache = mocker.MagicMock(QueryCache)
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    with patch(
        "servicex.servicex_adapter.ServiceXAdapter.get_code_generators",
        return_value=codegen_list,
    ):
        client = ServiceXClient(
            backend="servicex-uc-af", config_path="tests/example_config.yaml"
        )
        client.servicex = sx
        client.query_cache = mock_cache

        datasource = client.generic_query(
            dataset_identifier=did, query=UprootRawQuery({"treename": "CollectionTree"})
        )
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        _ = await datasource.submit_and_download(
            signed_urls_only=False, expandable_progress=progress
        )

    # same thing but a list argument to UprootRawQuery (UprootRawQuery test...)
    datasource = client.generic_query(
        dataset_identifier=did, query=UprootRawQuery({"treename": "CollectionTree"})
    )
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        _ = await datasource.submit_and_download(
            signed_urls_only=False, expandable_progress=progress
        )
    assert isinstance(json.loads(datasource.generate_selection_string()), list)


@pytest.mark.asyncio
async def test_submit_cancelled(mocker, codegen_list):
    """Uses Uproot-Raw classes which go through the query cancelled mechanism"""
    import json

    sx = AsyncMock()
    sx.submit_transform = AsyncMock()
    sx.submit_transform.return_value = {"request_id": '123-456-789"'}
    sx.get_transform_status = AsyncMock()
    sx.get_transform_status.side_effect = [transform_status4]

    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(side_effect=[[file1], [file1, file2]])
    mock_minio.download_file = AsyncMock()

    mock_cache = mocker.MagicMock(QueryCache)
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")
    with patch(
        "servicex.servicex_adapter.ServiceXAdapter.get_code_generators",
        return_value=codegen_list,
    ):
        client = ServiceXClient(
            backend="servicex-uc-af", config_path="tests/example_config.yaml"
        )
        client.servicex = sx
        client.query_cache = mock_cache

        datasource = client.generic_query(
            dataset_identifier=did, query=UprootRawQuery({"treename": "CollectionTree"})
        )
    with ExpandableProgress(display_progress=False) as progress:
        datasource.result_format = ResultFormat.parquet
        _ = await datasource.submit_and_download(
            signed_urls_only=False, expandable_progress=progress
        )
    assert isinstance(json.loads(datasource.generate_selection_string()), list)


def test_transform_request():
    servicex = AsyncMock()
    did = FileListDataset("/foo/bar/baz.root")
    with tempfile.TemporaryDirectory() as temp_dir:
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        datasource = Query(
            dataset_identifier=did,
            title="ServiceX Client",
            codegen="uproot",
            sx_adapter=servicex,
            query_cache=None,
            config=Configuration(api_endpoints=[]),
        )
        datasource.query_string_generator = (
            FuncADLQuery_Uproot()
            .FromTree("nominal")
            .Select(lambda e: {"lep_pt": e["lep_pt"]})
        )

        q = datasource.set_result_format(ResultFormat.parquet).transform_request
        assert (
            q.selection == "(call Select (call EventDataset 'bogus.root' 'nominal') "
            "(lambda (list e) (dict (list 'lep_pt') "
            "(list (subscript e 'lep_pt')))))"
        )
        cache.close()


@pytest.mark.asyncio
async def test_use_of_ignore_cache(mocker, servicex):
    """Run a normal request -> run the same request with ignore cache -> run request with cache
    After the first request the request is cached
    After the second request, transformer runs again by ignoring cache
    After the third request, the data is retrieved from the cache
    """
    # Prepare ServiceX
    servicex.get_transform_status.side_effect = cycle(
        [
            transform_status1,
            transform_status3,
        ]
    )

    # Prepare Minio
    mock_minio = AsyncMock()
    mock_minio.list_bucket = AsyncMock(return_value=[file1, file2])
    mock_minio.get_signed_url = AsyncMock(side_effect=["http://file1", "http://file2"])
    mocker.patch("servicex.minio_adapter.MinioAdapter", return_value=mock_minio)
    did = FileListDataset("/foo/bar/baz.root")

    with tempfile.TemporaryDirectory() as temp_dir:

        # Datasource without cache ignore
        config = Configuration(cache_path=temp_dir, api_endpoints=[])
        cache = QueryCache(config)
        datasource_without_ignore_cache = Query(
            dataset_identifier=did,
            title="ServiceX Client",
            codegen="uproot",
            sx_adapter=servicex,
            query_cache=cache,
            config=config,
        )
        datasource_without_ignore_cache.query_string_generator = (
            FuncADLQuery_Uproot().FromTree("nominal")
        )
        datasource_without_ignore_cache.result_format = ResultFormat.parquet

        # Datasouce with ignore cache
        datasource_with_ignore_cache = Query(
            dataset_identifier=did,
            title="ServiceX Client",
            codegen="uproot",
            sx_adapter=servicex,
            query_cache=cache,
            config=config,
            ignore_cache=True,
        )
        datasource_with_ignore_cache.query_string_generator = (
            FuncADLQuery_Uproot().FromTree("nominal")
        )
        datasource_with_ignore_cache.result_format = ResultFormat.parquet

        # 1st time sending the request
        upd = mocker.patch.object(
            cache, "update_record", side_effect=cache.update_record
        )
        with ExpandableProgress(display_progress=False) as progress:
            await datasource_without_ignore_cache.submit_and_download(
                signed_urls_only=True, expandable_progress=progress
            )  # noqa
        upd.assert_not_called()
        upd.reset_mock()
        assert mock_minio.get_signed_url.await_count == 2

        # 2nd time sending the same request with ignore_cache (So it will run again)
        mock_minio.get_signed_url = AsyncMock(
            side_effect=["http://file1", "http://file2"]
        )
        upd = mocker.patch.object(
            cache, "update_record", side_effect=cache.update_record
        )
        with ExpandableProgress(display_progress=False) as progress:
            await datasource_with_ignore_cache.submit_and_download(
                signed_urls_only=True, expandable_progress=progress
            )  # noqa
        upd.assert_not_called()
        upd.reset_mock()
        assert mock_minio.get_signed_url.await_count == 2

        # 3rd round, should hit the cache (and nothing else)
        servicex.get_transform_status.side_effect = [
            transform_status1,
            transform_status3,
        ]
        mock_minio.list_bucket.reset_mock()
        mock_minio.download_file.reset_mock()
        with ExpandableProgress(display_progress=False) as progress:
            res = await datasource_without_ignore_cache.submit_and_download(
                signed_urls_only=True, expandable_progress=progress
            )  # noqa
        mock_minio.list_bucket.assert_not_awaited()
        mock_minio.download_file.assert_not_awaited()
        assert len(res.signed_url_list) == 2
        cache.close()



from pathlib import Path

from servicex import ServiceXSpec
from servicex.servicex_client import _output_handler


def test_output_directory(tmp_path):
    config = {
        "General": {
            "Codegen": "python",
            "Delivery": "LocalCache",
            "OutputDirectory": str(tmp_path),
        },
        "Sample": [
            {"Name": "sampleA", "RucioDID": "user.kchoi:sampleA", "Query": "a"},
        ],
    }
    config = ServiceXSpec(**config)

    _output_handler(config, [], [])
    assert Path(tmp_path, "servicex_fileset.yaml").exists()



# Original code released under
# The MIT License (MIT)
#
#  Copyright (c) 2014-2018 Tristan Sweeney, Cambridge Consultants
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
# The code has been modified:
# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import types
from typing import Union, Any, Protocol
from pathlib import Path
import os

import ruamel.yaml
import ruamel.yaml.composer
import ruamel.yaml.constructor
from ruamel.yaml.nodes import ScalarNode, MappingNode, SequenceNode


class TextFileLike(Protocol):
    def read(self, size: int) -> Union[str, bytes]:
        """read function for a file-like object"""


class CompositingComposer(ruamel.yaml.composer.Composer):
    compositors = {k: {} for k in (ScalarNode, MappingNode, SequenceNode)}

    @classmethod
    def add_compositor(cls, tag, compositor, *, nodeTypes=(ScalarNode,)):
        for nodeType in nodeTypes:
            cls.compositors[nodeType][tag] = compositor

    @classmethod
    def get_compositor(cls, tag, nodeType):
        return cls.compositors[nodeType].get(tag, None)

    def __compose_dispatch(self, anchor, nodeType, callback):
        event = self.parser.peek_event()
        compositor = self.get_compositor(event.tag, nodeType) or callback
        if isinstance(compositor, types.MethodType):
            return compositor(anchor)
        else:
            return compositor(self, anchor)

    def compose_scalar_node(self, anchor):
        return self.__compose_dispatch(anchor, ScalarNode, super().compose_scalar_node)

    def compose_sequence_node(self, anchor):
        return self.__compose_dispatch(
            anchor, SequenceNode, super().compose_sequence_node
        )

    def compose_mapping_node(self, anchor):
        return self.__compose_dispatch(
            anchor, MappingNode, super().compose_mapping_node
        )


class ExcludingConstructor(ruamel.yaml.constructor.Constructor):
    filters = {k: [] for k in (MappingNode, SequenceNode)}

    @classmethod
    def add_filter(cls, filter, *, nodeTypes=(MappingNode,)):
        for nodeType in nodeTypes:
            cls.filters[nodeType].append(filter)

    def construct_mapping(self, node):
        node.value = [
            (key_node, value_node)
            for key_node, value_node in node.value
            if not any(f(key_node, value_node) for f in self.filters[MappingNode])
        ]
        return super().construct_mapping(node)

    def construct_sequence(self, node, deep=True):
        node.value = [
            value_node
            for value_node in node.value
            if not any(f(value_node) for f in self.filters[SequenceNode])
        ]
        return super().construct_sequence(node)


class YAML(ruamel.yaml.YAML):
    def __init__(self, *args, **kwargs):
        if "typ" not in kwargs:
            kwargs["typ"] = "safe"
        elif kwargs["typ"] not in ("safe", "unsafe") and kwargs["typ"] not in (
            ["safe"],
            ["unsafe"],
        ):  # pragma: no cover
            raise Exception(
                "Can't do typ={} parsing w/ composition time directives!".format(
                    kwargs["typ"]
                )
            )

        if "pure" not in kwargs:
            kwargs["pure"] = True
        elif not kwargs["pure"]:  # pragma: no cover
            raise Exception(
                "Can't do non-pure python parsing w/ composition time directives!"
            )

        super().__init__(*args, **kwargs)
        self.Composer = CompositingComposer
        self.Constructor = ExcludingConstructor

    def compose(self, stream: Union[Path, str, bytes, TextFileLike]) -> Any:
        """
        at this point you either have the non-pure Parser (which has its own reader and
        scanner) or you have the pure Parser.
        If the pure Parser is set, then set the Reader and Scanner, if not already set.
        If either the Scanner or Reader are set, you cannot use the non-pure Parser,
            so reset it to the pure parser and set the Reader resp. Scanner if necessary
        """
        constructor, parser = self.get_constructor_parser(stream)
        try:
            return self.composer.get_single_node()
        finally:
            parser.dispose()
            try:
                self._reader.reset_reader()
            except AttributeError:  # pragma: no cover
                pass
            try:
                self._scanner.reset_scanner()
            except AttributeError:  # pragma: no cover
                pass

    def fork(self):
        return type(self)(typ=self.typ, pure=self.pure)


def include_compositor(self, anchor):
    event = self.parser.get_event()
    yaml = self.loader.fork()
    path = os.path.join(os.path.dirname(self.loader.reader.name), event.value)
    with open(os.path.abspath(path)) as f:
        rv = yaml.compose(f)
        self.loader.composer.anchors.update(yaml.composer.anchors)
        return rv


def exclude_filter(key_node, value_node=None):
    value_node = value_node or key_node  # copy ref if None
    return key_node.tag == "!exclude" or value_node.tag == "!exclude"


CompositingComposer.add_compositor("!include", include_compositor)
ExcludingConstructor.add_filter(exclude_filter, nodeTypes=(MappingNode, SequenceNode))



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
import os
from pathlib import Path
from typing import List, Optional
from filelock import FileLock
from tinydb import TinyDB, Query, where

from servicex.configuration import Configuration
from servicex.models import TransformRequest, TransformStatus, TransformedResults


class CacheException(Exception):
    pass


class QueryCache:
    def __init__(self, config: Configuration):
        self.config = config
        Path(self.config.cache_path).mkdir(parents=True, exist_ok=True)
        self.db = TinyDB(os.path.join(self.config.cache_path, "db.json"))
        self.lock = FileLock(os.path.join(self.config.cache_path, "db.lock"))

    def close(self):
        self.db.close()

    def transformed_results(
        self,
        transform: TransformRequest,
        completed_status: TransformStatus,
        data_dir: str,
        file_list: List[str],
        signed_urls,
    ) -> TransformedResults:
        return TransformedResults(
            hash=transform.compute_hash(),
            title=transform.title,
            codegen=transform.codegen,
            request_id=completed_status.request_id,
            submit_time=completed_status.submit_time,
            data_dir=data_dir,
            file_list=file_list,
            signed_url_list=signed_urls,
            files=completed_status.files,
            result_format=transform.result_format,
            log_url=completed_status.log_url,
        )

    def cache_transform(self, record: TransformedResults):
        transforms = Query()
        with self.lock:
            self.db.upsert(
                json.loads(record.model_dump_json()), transforms.hash == record.hash
            )

    def update_record(self, record: TransformedResults):
        transforms = Query()
        with self.lock:
            self.db.update(
                json.loads(record.model_dump_json()), transforms.hash == record.hash
            )

    def contains_hash(self, hash: str) -> bool:
        """
        Check if the cache has completed records for a hash
        """
        transforms = Query()
        with self.lock:
            records = self.db.search(
                (transforms.hash == hash) & ~(transforms.status == "SUBMITTED")
            )
        return len(records) > 0

    def is_transform_request_submitted(self, hash_value: str) -> bool:
        """
        Returns True if request is submitted
        Returns False if the request is not in the cache at all
        or not submitted
        """
        transform = Query()
        with self.lock:
            records = self.db.search((transform.hash == hash_value))

        if not records:
            return False

        if "status" in records[0] and records[0]["status"] == "SUBMITTED":
            return True
        return False

    def get_transform_request_id(self, hash_value: str) -> Optional[str]:
        """
        Return the request id of cached record
        """
        transform = Query()

        with self.lock:
            records = self.db.search(transform.hash == hash_value)

        if not records or "request_id" not in records[0]:
            raise CacheException("Request Id not found")
        return records[0]["request_id"]

    def update_transform_status(self, hash_value: str, status: str) -> None:
        """
        Update the cached record status
        """
        transform = Query()
        with self.lock:
            self.db.upsert(
                {"hash": hash_value, "status": status}, transform.hash == hash_value
            )

    def update_transform_request_id(self, hash_value: str, request_id: str) -> None:
        """
        Update the cached record request id
        """
        transform = Query()
        with self.lock:
            self.db.upsert(
                {"hash": hash_value, "request_id": request_id},
                transform.hash == hash_value,
            )

    def get_transform_by_hash(self, hash: str) -> Optional[TransformedResults]:
        """
        Returns completed transformations by hash
        """
        transforms = Query()
        with self.lock:
            records = records = self.db.search(
                (transforms.hash == hash) & ~(transforms.status == "SUBMITTED")
            )

        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for hash")
        else:
            return TransformedResults(**records[0])

    def get_transform_by_request_id(
        self, request_id: str
    ) -> Optional[TransformedResults]:
        """
        Returns completed transformed results using a request id
        """
        transforms = Query()

        with self.lock:
            records = self.db.search(transforms.request_id == request_id)

        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for request_id")
        else:
            return TransformedResults(**records[0])

    def cache_path_for_transform(self, transform_status: TransformStatus) -> Path:
        base = Path(self.config.cache_path)
        result = Path(os.path.join(base, transform_status.request_id))
        result.mkdir(parents=True, exist_ok=True)
        return result

    def cached_queries(self) -> List[TransformedResults]:
        transforms = Query()

        with self.lock:
            result = [
                TransformedResults(**doc)
                for doc in self.db.search(transforms.request_id.exists())
            ]
        return result

    def delete_record_by_request_id(self, request_id: str):
        with self.lock:
            self.db.remove(where("request_id") == request_id)

    def delete_record_by_hash(self, hash: str):
        transforms = Query()
        with self.lock:
            self.db.remove(transforms.hash == hash)

    def get_codegen_by_backend(self, backend: str) -> Optional[dict]:
        codegens = Query()
        with self.lock:
            records = self.db.search(codegens.backend == backend)

        if not records:
            return None

        if len(records) != 1:
            raise CacheException("Multiple records found in db for same backend")
        else:
            return records[0]

    def update_codegen_by_backend(self, backend: str, codegen_list: list):
        transforms = Query()
        with self.lock:
            self.db.upsert(
                {"backend": backend, "codegens": codegen_list},
                transforms.backend == backend,
            )

    def delete_codegen_by_backend(self, backend: str):
        with self.lock:
            self.db.remove(where("backend") == backend)



# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
from typing import Optional

import rich

from servicex.app import pipeable_table, is_terminal_output
from servicex.app.cli_options import backend_cli_option, config_file_option

import typer

from servicex.servicex_client import ServiceXClient
from rich.table import Table

datasets_app = typer.Typer(name="datasets", no_args_is_help=True)
did_finder_opt = typer.Option(
    None,
    help="Filter datasets by DID finder. Some useful values are 'rucio' or 'user'",
    show_default=False,
)
show_deleted_opt = typer.Option(
    False,
    help="Show deleted datasets",
    show_default=True,
)
dataset_id_get_arg = typer.Argument(..., help="The ID of the dataset to get")
dataset_id_delete_arg = typer.Argument(..., help="The ID of the dataset to delete")


@datasets_app.command(no_args_is_help=False)
def list(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    did_finder: Optional[str] = did_finder_opt,
    show_deleted: Optional[bool] = show_deleted_opt,
):
    """
    List the datasets. Use fancy formatting if printing to a terminal.
    Output as plain text if redirected.
    """
    sx = ServiceXClient(backend=backend, config_path=config_path)
    table = pipeable_table(title="ServiceX Datasets")
    table.add_column("ID")
    table.add_column("Name")
    table.add_column("Files")
    table.add_column("Size")
    table.add_column("Status")
    table.add_column("Created")
    if show_deleted:
        table.add_column("Deleted")

    datasets = asyncio.run(
        sx.get_datasets(did_finder=did_finder, show_deleted=show_deleted)
    )
    for d in datasets:
        # Format the CachedDataset object into a table row
        # The last_updated field is what we should be displaying, but that is
        # currently set to 1970-00-00 in the server and is never updated.
        # Stick with the last_used field until
        # https://github.com/ssl-hep/ServiceX/issues/906 is resolved
        d_name = d.name if d.did_finder != "user" else "File list"
        is_stale = "Yes" if d.is_stale else ""
        last_used = d.last_used.strftime("%Y-%m-%dT%H:%M:%S")
        table.add_row(
            str(d.id),
            d_name,
            "%d" % d.n_files,
            "{:,}MB".format(round(d.size / 1e6)),
            d.lookup_status,
            last_used,
            is_stale,
        )
    rich.print(table)


@datasets_app.command(no_args_is_help=True)
def get(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    dataset_id: int = dataset_id_get_arg,
):
    """
    Get the details of a dataset. Output as a pretty, nested table if printing to a terminal.
    Output as json if redirected.
    """
    sx = ServiceXClient(backend=backend, config_path=config_path)
    if is_terminal_output():
        table = Table(title=f"Dataset ID {dataset_id}")
        table.add_column("Paths")
    else:
        table = None

    dataset = asyncio.run(sx.get_dataset(dataset_id))

    if table:
        for file in dataset.files:
            sub_table = Table(title="")
            sub_table.add_column(f"File ID: {file.id}")
            for path in file.paths.split(","):
                sub_table.add_row(path)

            table.add_row(sub_table)
        # Set alternating row styles
        table.row_styles = ["", ""]
        rich.print(table)
    else:
        data = {
            "dataset": {
                "id": dataset.id,
                "name": dataset.name,
                "files": [
                    {"id": file.id, "paths": file.paths.split(",")}
                    for file in dataset.files
                ],
            }
        }
        rich.print_json(data=data)


@datasets_app.command(no_args_is_help=True)
def delete(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    dataset_id: int = dataset_id_delete_arg,
):
    sx = ServiceXClient(backend=backend, config_path=config_path)
    result = asyncio.run(sx.delete_dataset(dataset_id))
    if result:
        typer.echo(f"Dataset {dataset_id} deleted")
    else:
        typer.echo(f"Dataset {dataset_id} not found")
        raise typer.Abort()



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import asyncio
from pathlib import Path
from typing import Optional, List
import webbrowser
import re
from enum import Enum

import rich
import rich.box
import typer
from rich.progress import Progress

from servicex.app import pipeable_table
from servicex.app.cli_options import backend_cli_option, config_file_option
from servicex.minio_adapter import MinioAdapter
from servicex.models import Status, ResultFile
from servicex.servicex_client import ServiceXClient

transforms_app = typer.Typer(name="transforms", no_args_is_help=True)
complete_opt = typer.Option(
    None, "--complete", help="Only show successfully completed transforms"
)
running_opt = typer.Option(
    None, "--running", help="Only show transforms that are currently running"
)
transform_id_arg = typer.Argument(help="Transform ID")
local_dir_opt = typer.Option(".", "-d", help="Local dir to download to")
concurrency_opt = typer.Option(
    20, "--concurrency", help="Number of concurrent downloads"
)
log_level_opt = typer.Option(
    "ERROR", "-l", "--log-level", help="Level of Logs", case_sensitive=False
)
time_frame_opt = typer.Option(
    "month", "-f", "--time-frame", help="Time Frame", case_sensitive=False
)


@transforms_app.callback()
def transforms():
    """
    Sub-commands for interacting with transforms.
    """
    pass


@transforms_app.command(no_args_is_help=False)
def list(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    complete: Optional[bool] = complete_opt,
    running: Optional[bool] = running_opt,
):
    """
    List the transforms that have been run.
    """

    def transform_filter(status: Status) -> bool:
        if complete and status == Status.complete:
            return True
        if running and status == Status.running:
            return True
        if not complete and not running:
            return True
        return False

    sx = ServiceXClient(backend=backend, config_path=config_path)

    table = pipeable_table(title="ServiceX Transforms")
    table.add_column("Transform ID")
    table.add_column("Title")
    table.add_column("Status")
    table.add_column("Files")
    transforms = sx.get_transforms()
    for t in transforms:
        if transform_filter(t.status):
            table.add_row(t.request_id, t.title, t.status, str(t.files_completed))

    rich.print(table)


@transforms_app.command(no_args_is_help=True)
def files(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    transform_id: str = transform_id_arg,
):
    """
    List the files that were produced by a transform.
    """

    async def list_files(sx: ServiceXClient, transform_id: str) -> List[ResultFile]:
        transform = await sx.get_transform_status_async(transform_id)
        minio = MinioAdapter.for_transform(transform)
        return await minio.list_bucket()

    sx = ServiceXClient(backend=backend, config_path=config_path)
    result_files = asyncio.run(list_files(sx, transform_id))
    table = pipeable_table(title=f"Files from {transform_id}")
    table.add_column("filename")
    table.add_column("Size(Mb)")
    table.add_column("Filetype")
    for f in result_files:
        table.add_row(f.filename, "%.2f" % (f.size / 1e6), f.extension)
    rich.print(table)


@transforms_app.command(no_args_is_help=True)
def download(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    transform_id: str = transform_id_arg,
    local_dir: str = local_dir_opt,
    concurrency: int = concurrency_opt,
):
    """
    Download the files that were produced by a transform.
    """

    async def download_files(sx: ServiceXClient, transform_id: str, local_dir):
        s3_semaphore = asyncio.Semaphore(concurrency)

        async def download_with_progress(filename) -> Path:
            async with s3_semaphore:
                p = await minio.download_file(
                    filename,
                    local_dir,
                    shorten_filename=sx.config.shortened_downloaded_filename,
                )
            progress.advance(download_progress)
            return p

        transform = await sx.get_transform_status_async(transform_id)
        minio = MinioAdapter.for_transform(transform)
        file_list = await minio.list_bucket()
        progress.update(download_progress, total=len(file_list))
        progress.start_task(download_progress)

        tasks = [download_with_progress(f.filename) for f in file_list]
        return await asyncio.gather(*tasks)

    with Progress() as progress:
        download_progress = progress.add_task("Downloading", start=False, total=None)
        sx = ServiceXClient(backend=backend, config_path=config_path)
        result_files = asyncio.run(download_files(sx, transform_id, local_dir))

    for path in result_files:
        print(path.as_posix())


@transforms_app.command(no_args_is_help=True)
def delete(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    transform_id_list: List[str] = transform_id_arg,
):
    """
    Delete a completed transform along with the result files.
    """
    sx = ServiceXClient(backend=backend, config_path=config_path)
    for transform_id in transform_id_list:
        asyncio.run(sx.delete_transform(transform_id))
        sx.delete_transform_from_cache(transform_id)

        print(f"Transform {transform_id} deleted")


@transforms_app.command(no_args_is_help=True)
def cancel(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    transform_id_list: List[str] = transform_id_arg,
):
    """
    Cancel a running transform request.
    """
    sx = ServiceXClient(backend=backend, config_path=config_path)
    for transform_id in transform_id_list:
        asyncio.run(sx.cancel_transform(transform_id))
        print(f"Transform {transform_id} cancelled")


class TimeFrame(str, Enum):
    r"""
    Time Frame levels: 'day', 'week' & 'month'
    """

    day = ("day",)
    week = ("week",)
    month = ("month",)


class LogLevel(str, Enum):
    r"""
    Level of the log messages: INFO & ERROR
    """

    info = ("INFO",)
    error = ("ERROR",)


def add_query(key, value):
    """
    Creates query string from the key and value pairs
    """
    query_string = "(query:(match_phrase:({0}:'{1}')))".format(key, value)
    return query_string


def select_time(time_frame=TimeFrame.day):
    """
    Takes input as 'day','week','month' and returns the time filter
    """
    time_string = time_frame
    if time_frame.lower() == TimeFrame.day:
        time_string = "time:(from:now%2Fd,to:now%2Fd)"
    elif time_frame.lower() == TimeFrame.week:
        time_string = "time:(from:now%2Fw,to:now%2Fw)"
    elif time_frame.lower() == TimeFrame.month:
        time_string = "time:(from:now-30d%2Fd,to:now)"
    else:
        rich.print("Got a time frame apart from 'day', 'week', 'month'")
    return time_string


def create_kibana_link_parameters(
    log_url, transform_id=None, log_level=None, time_frame=None
):
    """
    Create the _a and _g parameters for the kibana dashboard link
    """
    if log_level:
        a_parameter = (
            f"&_a=(filters:!({add_query('requestId', transform_id)},"
            f"{add_query('level', log_level.value.lower())}))"
        )
    else:
        a_parameter = f"&_a=(filters:!({add_query('requestId', transform_id)}))"
    g_parameter = f"&_g=({select_time(time_frame.value.lower())})"
    kibana_link = re.sub(r"\&\_g\=\(\)", g_parameter + a_parameter, log_url)
    return kibana_link


@transforms_app.command(no_args_is_help=True)
def logs(
    backend: Optional[str] = backend_cli_option,
    transform_id: str = transform_id_arg,
    log_level: Optional[LogLevel] = log_level_opt,
    time_frame: Optional[TimeFrame] = time_frame_opt,
):
    """
    Open the URL to the Kibana dashboard of the logs of a tranformer
    """
    sx = ServiceXClient(backend=backend)
    transforms = sx.get_transform_status(transform_id)
    if transforms and transforms.request_id == transform_id:
        kibana_link = create_kibana_link_parameters(
            transforms.log_url,
            transform_id=transform_id,
            log_level=log_level,
            time_frame=time_frame,
        )
        print(kibana_link)
        webbrowser.open(kibana_link)
    else:
        rich.print("Invalid Request ID")



# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from servicex.func_adl.func_adl_dataset import FuncADLQuery_Uproot, FuncADLQuery
from typing import Any
import pytest


def test_set_from_tree():
    query = FuncADLQuery_Uproot()
    query = query.FromTree("TREE_NAME")

    assert "TREE_NAME" in query.generate_selection_string()


def test_requires_tree():
    query = FuncADLQuery_Uproot()

    with pytest.raises(ValueError):
        query.generate_selection_string()


def test_a_query():
    query = FuncADLQuery_Uproot()
    query = query.FromTree("nominal").Select(lambda e: {"lep_pt": e["lep_pt"]})

    assert (
        query.generate_selection_string()
        == "(call Select (call EventDataset 'bogus.root' 'nominal') "
        "(lambda (list e) (dict (list 'lep_pt') "
        "(list (subscript e 'lep_pt')))))"
    )


def test_set_query():
    qastle = (
        "(call Select (call EventDataset 'bogus.root' 'nominal') "
        "(lambda (list e) (dict (list 'lep_pt') "
        "(list (subscript e 'lep_pt')))))"
    )
    query = FuncADLQuery_Uproot()
    query.set_provided_qastle(qastle)

    assert query.generate_selection_string() == qastle


def test_type():
    "Test that the item type for a dataset is correctly propagated"

    class my_type_info:
        "typespec for possible event type"

        def fork_it_over(self) -> int: ...

    datasource = FuncADLQuery[my_type_info](item_type=my_type_info)

    assert datasource.item_type == my_type_info


def test_type_any():
    "Test the type is any if no type is given"
    datasource = FuncADLQuery()
    assert datasource.item_type == Any



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import hashlib
from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field, field_validator
from typing import List, Optional, Any


def _get_typename(typeish) -> str:
    return typeish.__name__ if isinstance(typeish, type) else str(typeish)


def _generate_model_docstring(model: type) -> str:
    NL = "\n"
    return "\n".join(
        [(model.__doc__ if model.__doc__ else model.__name__).strip(), "", "Args:"]
        + [
            f"    {field}: ({_get_typename(info.annotation)}) "
            f'{info.description.replace(NL, NL + " " * 8) if info.description else ""}'
            for field, info in model.model_fields.items()
        ]
    )


class DocStringBaseModel(BaseModel):
    """Class to autogenerate a docstring for a Pydantic model"""

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any):
        super().__pydantic_init_subclass__(**kwargs)
        # There is currently no good way of knowing we are building within Sphinx.
        # Use a hacky workaround but monitor https://github.com/sphinx-doc/sphinx/issues/9805
        import os

        if "IN_SPHINX_BUILD" not in os.environ:
            cls.__doc__ = _generate_model_docstring(cls)


class ResultDestination(str, Enum):
    r"""
    Direct the output to object store or posix volume
    """

    object_store = "object-store"
    volume = "volume"


class ResultFormat(str, Enum):
    r"""
    Specify the file format for the generated output
    """

    parquet = "parquet"
    root_ttree = "root-file"


class Status(str, Enum):
    r"""
    Status of a submitted transform
    """

    complete = ("Complete",)
    fatal = ("Fatal",)
    canceled = ("Canceled",)
    submitted = ("Submitted",)
    looking = ("Lookup",)
    pending = "Pending Lookup"
    running = "Running"


class TransformRequest(DocStringBaseModel):
    r"""
    Transform request sent to ServiceX
    """

    title: Optional[str] = None
    did: Optional[str] = None
    file_list: Optional[List[str]] = Field(default=None, alias="file-list")
    selection: str
    codegen: str
    result_destination: ResultDestination = Field(
        serialization_alias="result-destination"
    )
    result_format: ResultFormat = Field(serialization_alias="result-format")

    model_config = {"populate_by_name": True, "use_attribute_docstrings": True}

    def compute_hash(self):
        r"""
        Compute a hash for this submission. Only include properties that impact the result
        so we have maximal ability to reuse transforms

        :return: SHA256 hash of request
        """
        sha = hashlib.sha256(
            str(
                [
                    self.did,
                    self.selection,
                    None,  # was tree_name
                    self.codegen,
                    None,  # was image
                    self.result_format.name,
                    sorted(self.file_list) if self.file_list else None,
                ]
            ).encode("utf-8")
        )
        return sha.hexdigest()


class TransformStatus(DocStringBaseModel):
    r"""
    Status object returned by servicex
    """

    model_config = {"use_attribute_docstrings": True}

    request_id: str
    did: str
    title: Optional[str] = None
    selection: str
    tree_name: Optional[str] = Field(validation_alias="tree-name")
    image: str
    result_destination: ResultDestination = Field(validation_alias="result-destination")
    result_format: ResultFormat = Field(validation_alias="result-format")
    generated_code_cm: str = Field(validation_alias="generated-code-cm")
    status: Status
    app_version: str = Field(validation_alias="app-version")
    files: int
    files_completed: int = Field(validation_alias="files-completed")
    files_failed: int = Field(validation_alias="files-failed")
    files_remaining: Optional[int] = Field(
        validation_alias="files-remaining", default=0
    )
    submit_time: datetime = Field(validation_alias="submit-time", default=None)
    finish_time: Optional[datetime] = Field(
        validation_alias="finish-time", default=None
    )
    minio_endpoint: Optional[str] = Field(
        validation_alias="minio-endpoint", default=None
    )
    minio_secured: Optional[bool] = Field(
        validation_alias="minio-secured", default=None
    )
    minio_access_key: Optional[str] = Field(
        validation_alias="minio-access-key", default=None
    )
    minio_secret_key: Optional[str] = Field(
        validation_alias="minio-secret-key", default=None
    )
    log_url: Optional[str] = Field(validation_alias="log-url", default=None)

    @field_validator("finish_time", mode="before")
    @classmethod
    def parse_finish_time(cls, v):
        if isinstance(v, str) and v == "None":
            return None
        return v


class ResultFile(DocStringBaseModel):
    r"""
    Record reporting the properties of a transformed file result
    """

    model_config = {"use_attribute_docstrings": True}

    filename: str
    size: int
    extension: str


class TransformedResults(DocStringBaseModel):
    r"""
    Returned for a submission. Gives you everything you need to know about a completed
    transform.
    """

    model_config = {"use_attribute_docstrings": True}

    hash: str
    """Unique hash for transformation (used to look up results in cache)"""
    title: str
    """Title of transformation request"""
    codegen: str
    """Code generator used (internal ServiceX information related to query type)"""
    request_id: str
    """Associated request ID from the ServiceX server"""
    submit_time: datetime
    """Time of submission"""
    data_dir: str
    """Local directory for output"""
    file_list: List[str]
    """List of downloaded files on local disk"""
    signed_url_list: List[str]
    """List of URLs to retrieve output from remote ServiceX object store"""
    files: int
    """Number of files in result"""
    result_format: ResultFormat
    """File format for results"""
    log_url: Optional[str] = None
    """URL for looking up logs on the ServiceX server"""


class DatasetFile(BaseModel):
    """
    Model for a file in a cached dataset
    """

    id: int
    adler32: Optional[str]
    file_size: int
    file_events: int
    paths: str


class CachedDataset(BaseModel):
    """
    Model for a cached dataset held by ServiceX server
    """

    id: int
    name: str
    did_finder: str
    n_files: int
    size: int
    events: int
    last_used: datetime
    last_updated: datetime
    lookup_status: str
    is_stale: bool
    files: Optional[List[DatasetFile]] = []



# Copyright (c) 2024, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from enum import Enum
import hashlib
from typing import Union, Optional, List
from pydantic import (
    Field,
    model_validator,
    field_validator,
)
import logging

from servicex.dataset_identifier import (
    DataSetIdentifier,
    RucioDatasetIdentifier,
    FileListDataset,
)
from servicex.query_core import QueryStringGenerator
from servicex.models import ResultFormat, DocStringBaseModel

logger = logging.getLogger(__name__)


class Sample(DocStringBaseModel):
    """
    Represents a single transform request within a larger submission.
    """

    model_config = {"use_attribute_docstrings": True}

    Name: str
    """
    The name of the sample. This makes it easier to identify the sample
    in the output.
    """

    Dataset: Optional[DataSetIdentifier] = None
    """
    Dataset identifier for the sample
    """

    NFiles: Optional[int] = Field(default=None)
    """
    Limit the Number of files to be used in the sample.
    The DID Finder will guarantee the same files will be returned
    between each invocation. Set to `None` to use all files.
    """

    Query: Optional[Union[str, QueryStringGenerator]] = Field(default=None)
    """
    Query string or query generator for the sample.
    """

    IgnoreLocalCache: bool = False
    """
    Flag to ignore local cache.
    """

    Codegen: Optional[str] = None
    """
    Code generator name, if applicable. Generally users don't need to specify this. It is
    implied by the query class
    """

    RucioDID: Optional[str] = Field(default=None, deprecated="Use 'Dataset' instead.")
    """
    Rucio Dataset Identifier, if applicable.
        Deprecated: Use 'Dataset' instead.
    """

    XRootDFiles: Optional[Union[str, List[str]]] = Field(default=None, deprecated=True)
    """
    XRootD file(s) associated with the sample.
            Deprecated: Use 'Dataset' instead.
    """

    model_config = {"arbitrary_types_allowed": True}

    @property
    def dataset_identifier(self) -> DataSetIdentifier:
        """
        Access the dataset identifier for the sample.
        """
        if self.Dataset:
            if self.NFiles is not None:
                self.Dataset.num_files = self.NFiles
            return self.Dataset
        elif self.RucioDID:
            return RucioDatasetIdentifier(self.RucioDID, num_files=self.NFiles)
        elif self.XRootDFiles:
            return FileListDataset(self.XRootDFiles)
        else:  # pragma: no cover
            raise RuntimeError("No valid dataset found, somehow validation failed")

    @model_validator(mode="before")
    @classmethod
    def validate_did_xor_file(cls, values):
        """
        Ensure that only one of Dataset, RootFile, or RucioDID is specified.
        :param values:
        :return:
        """
        count = sum(
            ["RucioDID" in values, "XRootDFiles" in values, "Dataset" in values]
        )
        if count > 1:
            raise ValueError("Only specify one of Dataset, XRootDFiles, or RucioDID.")
        if count == 0:
            raise ValueError("Must specify one of Dataset, XRootDFiles, or RucioDID.")
        return values

    @model_validator(mode="after")
    def validate_nfiles_is_not_zero(self):
        """
        Ensure that NFiles is not set to zero
        """
        if self.dataset_identifier.num_files == 0:
            raise ValueError("NFiles cannot be set to zero for a dataset.")
        return self

    @field_validator("Name", mode="before")
    @classmethod
    def truncate_long_sample_name(cls, v):
        """
        Truncate sample name to 128 characters if exceed
        Print warning message
        """
        if len(v) > 128:
            logger.warning(f"Truncating Sample name to 128 characters for {v}")
            v = v[0:128]
        return v

    @property
    def hash(self):
        sha = hashlib.sha256(
            str(
                [
                    self.dataset_identifier.hash,
                    self.NFiles,
                    (
                        self.Query
                        if (not self.Query or isinstance(self.Query, str))
                        else self.Query.generate_selection_string()
                    ),
                    self.Codegen,
                ]
            ).encode("utf-8")
        )
        return sha.hexdigest()


class General(DocStringBaseModel):
    """
    Represents a group of samples to be transformed together.
    """

    model_config = {"use_attribute_docstrings": True}

    class OutputFormatEnum(str, Enum):
        """
        Specifies the output format for the transform request.
        """

        parquet = "parquet"
        """
        Save the output as
        a parquet file https://parquet.apache.org/
        """

        root_ttree = "root-ttree"
        """
        Save the output as
        a ROOT TTree https://root.cern.ch/doc/master/classTTree.html
        """

        def to_ResultFormat(self) -> ResultFormat:
            """This method is used to convert the OutputFormatEnum enum to the ResultFormat enum,
            which is what is actually used for the TransformRequest. This allows us to use
            different string values in the two enum classes to maintain backend compatibility
            """
            if self == self.parquet:
                return ResultFormat.parquet
            elif self == self.root_ttree:
                return ResultFormat.root_ttree
            else:  # pragma: no cover
                raise RuntimeError(f"Bad OutputFormatEnum {self}")

    class DeliveryEnum(str, Enum):
        LocalCache = "LocalCache"
        """
        Download the files to the local computer and store them in the cache.
        Transform requests will return paths to these files in the cache
        """

        URLs = "URLs"
        """
        Return URLs to the files stored in the ServiceX object store
        """

    Codegen: Optional[str] = None
    """
    Code generator name to be applied across all of the samples, if applicable.
    Generally users don't need to specify this. It is implied by the query class
    """
    OutputFormat: OutputFormatEnum = Field(default=OutputFormatEnum.root_ttree)
    """
    Output format for the transform request.
    """

    Delivery: DeliveryEnum = Field(default=DeliveryEnum.LocalCache)
    """
    Specifies the delivery method for the output files.
    """

    OutputDirectory: Optional[str] = None
    """
    Directory to output a yaml file describing the output files.
    """

    OutFilesetName: str = "servicex_fileset"
    """
    Name of the yaml file that will be created in the output directory.
    """

    IgnoreLocalCache: bool = False
    """
    Flag to ignore local cache for all samples.
    """


# TODO: ServiceXSpec class has a field name General and it clashes with the class name General
# when it is called General() to initialize default values for General class
_General = General


class ServiceXSpec(DocStringBaseModel):
    """
    ServiceX Submission Specification - pass this into the ServiceX `deliver` function
    """

    model_config = {"use_attribute_docstrings": True}

    General: _General = General()
    """
    General settings for the transform request
    """

    Sample: List[Sample]
    """
    List of samples to be transformed
    """

    Definition: Optional[List] = None
    """
    Any reusable definitions that are needed for the transform request
    """

    @field_validator("Sample", mode="after")
    @classmethod
    def validate_unique_sample(cls, v):
        hash_set = set()
        for sample in v:
            if sample.hash in hash_set:
                raise RuntimeError(f"Duplicate samples detected: {sample.Name}")
            hash_set.add(sample.hash)
        return v



# Copyright (c) 2022, IRIS-HEP
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional

import rich
import typer

from servicex import servicex_client
from servicex._version import __version__
from servicex.app.cli_options import backend_cli_option, config_file_option
from servicex.app.datasets import datasets_app
from servicex.app.transforms import transforms_app
from servicex.app.cache import cache_app
from servicex.app.codegen import codegen_app

app = typer.Typer(no_args_is_help=True)

app.add_typer(transforms_app)
app.add_typer(cache_app)
app.add_typer(codegen_app)
app.add_typer(datasets_app)

spec_file_arg = typer.Argument(..., help="Spec file to submit to serviceX")
ignore_cache_opt = typer.Option(
    None, "--ignore-cache", help="Ignore local cache and always submit to ServiceX"
)


def show_version(show: bool):
    """Display the installed version and quit."""
    if show:
        rich.print(f"ServiceX {__version__}")
        raise typer.Exit()


version_opt = typer.Option(None, "--version", callback=show_version, is_eager=True)


@app.callback()
def main_info(version: Optional[bool] = version_opt):
    """
    ServiceX Client
    """
    pass


@app.command()
def deliver(
    backend: Optional[str] = backend_cli_option,
    config_path: Optional[str] = config_file_option,
    spec_file: str = spec_file_arg,
    ignore_cache: Optional[bool] = ignore_cache_opt,
):
    """
    Deliver a file to the ServiceX cache.
    """

    print(f"Delivering {spec_file} to ServiceX cache")
    results = servicex_client.deliver(
        spec_file,
        servicex_name=backend,
        config_path=config_path,
        ignore_local_cache=ignore_cache,
    )
    rich.print(results)


if __name__ == "__main__":
    app()
